/* (c) 2009 Richard Andrews */

#if !defined(UON_TRANSFORMS_H_)
#define UON_TRANSFORMS_H_

#include "uon_keyfile.h"
#include "uon_wire.h"


#define UON_TRANSFORM_ID_INVAL          0       /* marks uninitialised data */
#define UON_TRANSFORM_ID_NULL           1
#define UON_TRANSFORM_ID_TWOFISH        2
#define UON_TRANSFORM_ID_AESCBC         3
#define UON_TRANSFORM_ID_LZO            4
#define UON_TRANSFORM_ID_TWOFISH_LZO    5
#define UON_TRANSFORM_ID_AESCBC_LZO     6
#define UON_TRANSFORM_ID_USER_START     64
#define UON_TRANSFORM_ID_MAX            65535


struct uon_trans_op;
typedef struct uon_trans_op uon_trans_op_t;

struct uon_tostat {
    uint8_t             can_tx;         /* Does this transop have a valid SA for encoding. */
    uon_cipherspec_t    tx_spec;        /* If can_tx, the spec used to encode. */
};

typedef struct uon_tostat uon_tostat_t;


typedef int             (*uon_transdeinit_f)( uon_trans_op_t * arg );
typedef int             (*uon_transaddspec_f)( uon_trans_op_t * arg, 
                                               const uon_cipherspec_t * cspec );
typedef uon_tostat_t    (*uon_transtick_f)( uon_trans_op_t * arg, 
                                            time_t now );

typedef int             (*uon_transform_f)( uon_trans_op_t * arg,
                                            uint8_t * outbuf,
                                            size_t out_len,
                                            const uint8_t * inbuf,
                                            size_t in_len );

/** Holds the info associated with a data transform plugin.
 *
 *  When a packet arrives the transform ID is extracted. This defines the code
 *  to use to decode the packet content. The transform code then decodes the
 *  packet and consults its internal key lookup.
 */
struct uon_trans_op {
    void *              priv;   /* opaque data. Key schedule goes here. */

    uon_transform_t     transform_id;   /* link header enum to a transform */
    size_t              tx_cnt;
    size_t              rx_cnt;

    uon_transdeinit_f   deinit; /* destructor function */
    uon_transaddspec_f  addspec; /* parse opaque data from a key schedule file. */
    uon_transtick_f     tick;   /* periodic maintenance */
    uon_transform_f     fwd;    /* encode a payload */
    uon_transform_f     rev;    /* decode a payload */
};

/* Setup a single twofish SA for single-key operation. */
int transop_twofish_setup( uon_trans_op_t * ttt, 
                           uon_sa_t sa_num,
                           uint8_t * encrypt_pwd, 
                           uint32_t encrypt_pwd_len );

/* Initialise an empty transop ready to receive cipherspec elements. */
int  transop_twofish_init( uon_trans_op_t * ttt );
int  transop_aes_init( uon_trans_op_t * ttt );
void transop_null_init( uon_trans_op_t * ttt );

#endif /* #if !defined(UON_TRANSFORMS_H_) */

